/**
 * @file Interact with k-fet API.
 * @copyright 2017 cof-geek
 * @license MIT
 */


/**
 * Virtual namespace for models.
 * <br><br>
 *
 * A model subclasses {@link Models.ModelObject}.<br>
 * A model whose instances can be got from API subclasses
 * {@link Models.APIModelObject}.<br>
 * These two classes should not be used directly.
 * <br><br>
 *
 * Models with API support:
 * {@link Models.Account} (partial),
 * {@link Models.Checkout} (partial).
 * <br>
 * Models without API support:
 * {@link Models.Statement}.
 *
 * @namespace Models
 */


/**
 * Extended object
 * @memberof Models
 */
class ModelObject {

    /**
     * These properties always exist on instances of this class.
     * @abstract
     * @type {string[]}
     */
    static get props() { return []; }

    /**
     * Default values for properties given by <tt>props</tt> static member.
     * @abstract
     * @type {Object.<string, *>}
     */
    static get default_data() { return {}; }

    /**
     * Create new instance from data or default values.
     * @param   {Object} [data={}] - data to store in instance
     */
    constructor(data) {
        this.from(data || {});
    }

    /**
     * Check if an instance is empty.
     * @return  {boolean}
     */
    is_empty() {
        return ( this.id === undefined ) ? false : this.id == 0;
    }

    /**
     * Returns a {@link Formatters|Formatter} used for rendering.
     * @default {@link Formatters.Formatter}
     * @type {Formatter}
     */
    formatter() { return Formatter; }

    /**
     * Set properties of this instance from data object ones. If a property is
     * not present in data object, its value fallback to
     * {@link Models.ModelObject.default_data} one.
     * @param   {Object} data
     * @todo    Restrict to props
     */
    from(data) {
        // TODO: add restrict
        this.clear();
        this.update(data);
    }

    /**
     * Update properties of this instance from data ones.
     * @param   {Object} data
     */
    update(data) {
        $.extend(this, data);
    }

    /**
     * Clear properties to {@link Models.ModelObject.default_data|default_data}.
     */
    clear() {
        $.extend(this, this.constructor.default_data);
    }


    /**
     * Display stored data in container.
     * @param   {jQuery} $container
     * @param   {Object} [options] Options for formatter render method.
     * @param   {Formatters.Formatter}
     *          [formatter={@link Models.ModelObject#formatter|this.formatter()}]
     *          Formatter class to use.
     * @return  {jQuery} The DOM element $container, allowing methods chaining.
     */
    display($container, options, formatter) {
        formatter = formatter || this.formatter();
        formatter.render(this, $container, options);
        return $container;
    }

}


/**
 * Describes a model whose instances can be obtained through API.
 * @extends Models.ModelObject
 * @memberof Models
 */
class APIModelObject extends ModelObject {

    /**
     * Request url to get array of model instances data.
     * @abstract
     * @type {string}
     */
    static get url_model() {}

    /**
     * Request url to create an instance.
     * @abstract
     * @type {string}
     */
    static url_create() {}

    /**
     * Request url to edit an instance of this model.
     * @abstract
     * @param   {*} api_pk - Identifier of a model instance.
     * @type {string}
     */
    static url_update_for(api_pk) {}

    /**
     * Request url to get a single model instance data.
     * @abstract
     * @param   {*} api_pk - Identifier of a model instance.
     * @return  {string}
     */
    static url_read_for(api_pk) {}

    /**
     * See {@link Models.ModelObject|new ModelObject(data)}.
     * @param   {Object} [data={}] - data to store in instance
     */
    constructor(data) {
        super(data);

        if (this.id === undefined)
            this.id = 0;
    }

    /**
     * Identifier of the model instance in api requests.
     * @default <tt>this.id</tt>
     */
    get api_pk() { return this.id; }

    /**
     * Request url used to get current instance data.
     */
    get url_read() {
        if (this._url_object === undefined)
            return this.is_empty() ? '' : this.constructor.url_read_for(this.api_pk);
        return this._url_object;
    }

    set url_read(v) { this._url_object = v; }

    /**
     * Get data of a distant model instance. It sends a GET HTTP request to
     * {@link Models.APIModelObject#url_read}.
     * @param   {object} [api_options] Additional data appended to the request.
     */
    fromAPI(api_options) {
        api_options = api_options || {};

        api_options['format'] = 'json';

        return $.getJSON(this.url_read, api_options)
            .done( (json) => this.from(json) );
    }

    /**
     * Get data of a distant model instance via its id.
     * @param   {*} api_pk
     * @param   {object} [api_options]
     * @param   {jQueryAjaxSuccess} [on_success]
     * @param   {jQueryAjaxError} [on_error]
     * @see {@link Models.APIModelObject#fromAPI|fromAPI}
     */
    get_by_apipk(api_pk, api_options) {
        this.url_read = this.constructor.url_read_for(api_pk);
        return this.fromAPI(api_options);
    }

    /**
     * Add <tt>empty_props</tt> and <tt>empty_attrs</tt> options to
     * <tt>render</tt> call in {@link Models.ModelObject#display} if instance
     * is empty.
     * @see {@link Models.ModelObject#display}
     * @see {@link Models.APIModelObject#is_empty}
     */
    display($container, options, formatter) {
        if (this.is_empty()) {
            options.empty_props = true;
            options.empty_attrs = true;
        }
        return ModelObject.prototype.display.call(this, $container, options, formatter);
    }

}


/**
 * Account model. Can be accessed through API.
 * @extends Models.APIModelObject
 * @memberof Models
 */
class Account extends APIModelObject {

    /**
     * Properties retrieved through API.
     * @default <tt>['id', 'trigramme', 'name', 'nickname', 'email', 'is_cof',
     *          'promo', 'balance', 'is_frozen', 'departement']</tt>
     * @see {@link Models.ModelObject.props|ModelObject.props}
     */
    static get props() {
        return ['id', 'trigramme', 'name', 'nickname', 'email', 'is_cof',
                'promo', 'balance', 'is_frozen', 'departement'];
    }

    /**
     * Default values for Account model instances.
     * @default <tt>{ 'id': 0, 'trigramme': '', 'name': '', 'nickname': '',
     *          'email': '', ''is_cof': false, 'promo': '', 'balance': 0,
     *          'is_frozen': false, 'departement': '' }</tt>
     * @see {@link Models.ModelObject.default_data|ModelObject.default_data}
     */
    static get default_data() {
        return {
            'id': 0, 'trigramme': '', 'name': '', 'nickname': '', 'email': '',
            'is_cof' : false, 'promo': '', 'balance': 0, 'is_frozen': false,
            'departement': '',
        };
    };

    /**
     * @default <tt>django-js-reverse('kfet.account')</tt>
     * @see {@link Models.APIModelObject.url_model|APIModelObject.url_model}
     */
    static get url_model() { return Urls['kfet.account'](); }

    static url_create(trigramme) {
        var url = Urls['kfet.account.create']();
        if (trigramme) {
            var trigramme_url = encodeURIComponent(trigramme);
            url += `?trigramme=${trigramme_url}`
        }
        return url;
    }

    /**
     * @default <tt>django-js-reverse('kfet.account.read')(trigramme)</tt>
     * @param   {string} trigramme
     * @see {@link Models.APIModelObject.url_read_for|APIModelObject.url_read_for}
     */
    static url_read_for(trigramme) {
        var trigramme_url = encodeURIComponent(trigramme);
        return Urls['kfet.account.read'](trigramme_url);
    }

    /**
     * @default <tt>this.trigramme</tt>
     */
    get api_pk() { return this.trigramme; }

    /**
     * @default {@link Formatters.AccountFormatter}
     */
    formatter() {
        return (this.trigramme == 'LIQ') ? LIQFormatter : AccountFormatter;
    }

    // take care of "balance" type
    // API currently returns a string object (serialization of Decimal type within Django)
    get balance()  { return this._balance; }
    set balance(v) { return this._balance = floatCheck(v); }

    /**
     * Balance converted to UKF according to cof status.
     */
    get balance_ukf() { return amountToUKF(this.balance, this.is_cof); }

}


/**
 * Checkout model. Can be accessed through API.
 * @extends Models.APIModelObject
 * @memberof Models
 */
class Checkout extends APIModelObject {

    /**
     * Properties retrieved through API.
     * @default <tt>['id', 'name', 'balance', 'valid_from', 'valid_to']</tt>
     * @see {@link Models.ModelObject.props|ModelObject.props}
     */
    static get props() {
        return ['id', 'name', 'balance', 'valid_from', 'valid_to'];
    }

    /**
     * Default values for Account model instances.
     * @default <tt>{ 'id': 0, 'name': '', 'balance': 0, 'valid_from': '',
     *          'valid_to': '' }</tt>
     * @see {@link Models.ModelObject.default_data|ModelObject.default_data}
     */
    static get default_data() {
        return {
            'id': 0, 'name': '', 'balance': 0, 'valid_from': '', 'valid_to': '',
        };
    }

    /**
     * Not supported by API yet.
     * @see {@link Models.APIModelObject.url_model|APIModelObject.url_model}
     */
    static get url_model() { return ''; }

    /**
     * @default <tt>django-js-reverse('kfet.kpsul.checkout_data.read')(pk)</tt>
     * @param   {string} api_pk - a checkout id
     * @see {@link Models.APIModelObject.url_read_for|APIModelObject.url_read_for}
     */
    static url_read_for(api_pk) {
        return Urls['kfet.checkout.read'](api_pk);
    }

    /**
     * @default {@link Formatters.CheckoutFormatter}
     */
    formatter() {
        return CheckoutFormatter;
    }

    // take care of "balance" type
    // API currently returns a string object (serialization of Decimal type within Django)
    get balance()  { return this._balance; }
    set balance(v) { this._balance = floatCheck(v); }

}


/**
 * Statement model. Cannot be accessed through API.
 * @extends Models.ModelObject
 * @memberof Models
 */
class Statement extends ModelObject {

    /**
     * Properties associated to a statement.
     * @default <tt>['id', 'at', 'balance_old', 'balance_new', 'by']</tt>
     * @see {@link Models.ModelObject.props|ModelObject.props}
     */
    static get props() {
        return ['id', 'at', 'balance_old', 'balance_new', 'by'];
    }

    /**
     * Default values for Statement model instances.
     * @default <tt>{ 'id': 0, 'at': '', 'balance_old': 0, 'balance_new': 0,
     *          'by': '' }</tt>
     * @see {@link Models.ModelObject.default_data|ModelObject.default_data}
     */
    static get default_data() {
        return {
            'id': 0, 'at': '', 'balance_old': 0, 'balance_new': 0, 'by': '',
        };
    }

    static url_create(checkout_pk) {
        return Urls['kfet.checkoutstatement.create'](checkout_pk);
    }

    /**
     * @default {@link Formatters.StatementFormatter}
     */
    formatter() {
        return StatementFormatter;
    }

    // take care of "balance" type
    // API currently returns a string object (serialization of Decimal type within Django)
    get balance_old()  { return this._balance_old; }
    set balance_old(v) { this._balance_old = floatCheck(v); }

    get balance_new()  { return this._balance_new; }
    set balance_new(v) { this._balance_new = floatCheck(v); }

    get at()  { return this._at; }
    set at(v) { this._at = moment.isMoment(v) ? v : moment.tz(v, 'UTC'); }

}


/* ---------- ---------- */

/**
 * Virtual namespace for formatters.
 * <br><br>
 *
 * A formatter subclasses {@link Formatters.Formatter}.<br>
 * Formatters should be accessed statically only.
 *
 * @namespace Formatters
 */


/**
 * @memberof Formatters
 */
class Formatter {

    static get props() { return []; }
    static get attrs() { return []; }

    /**
     * Get value of a property through current formatter if defined, object itself otherwise.
     *
     * @param   {*} object - Object used for formatting
     * @param   {string} name - Name of property to render
     * @param   {string} [prefix=''] - Prefix used for formatter method search
     * @return  {*} Formatted property value
     */
    static get(object, name, prefix) {
        prefix = prefix || '';
        var method = prefix + name;
        if (this[method] !== undefined)
            return this[method](object);
        return object[name];
    }

    /**
     * Get formatted value of a property using prefix 'prop_'.
     * @see Formatter.get
     */
    static get_prop(object, name) {
        return this.get(object, name, 'prop_');
    }

    /**
     * Get formatted value of a property using prefix 'attr_'.
     * @see Formatter.get
     */
    static get_attr(obj, attr) {
        return this.get(obj, attr, 'attr_');
    }

    /**
     * Render container.
     * @param   {*} object - Object used for formatting
     * @param   {jQueryNode} $container
     * @param   {object} [options]
     *                   <tt>options['props']</tt> should be an array of
     *                   properties (default: <tt>class.props</tt>).<br>
     *                   <tt>options['prefix_prop']</tt> is added in front of
     *                   properties names in selector (default: <tt>'.'</tt>).<br>
     *                   <tt>options['prefix_attr']</tt> is added in front of
     *                   attributes names for container (default: <tt>''</tt>).
     */
    static render(object, $container, options) {
        options.props = options.props || [];
        options.attrs = options.attrs || [];

        var props = options.override_props ? options.props : this.props.concat(options.props);
        var attrs = options.override_attrs ? options.attrs : this.attrs.concat(options.attrs);

        var prefix_prop = options.prefix_prop !== undefined ? options.prefix_prop : '.';
        var prefix_attr = options.prefix_attr !== undefined ? options.prefix_attr : '';

        for (let prop of props) {
            var selector = prefix_prop + prop;
            var html = options.empty_props ? '' : this.get_prop(object, prop);
            $container.find( selector ).html( html );
        }

        for (let attr of attrs) {
            var name = prefix_attr + attr;
            var value = options.empty_attrs ? '' : this.get_attr(object, attr);
            $container.attr( name, value );
        }

        return $container;
    }

}


/**
 * @memberof Formatters
 * @extends Formatters.Formatter
 */
class AccountFormatter extends Formatter {

    /**
     * Properties renderable in html.
     * @default {@link Models.Account.props}<tt> + ['balance_ukf']</tt>
     */
    static get props() {
        return Account.props.concat(['balance_ukf']);
    }

    /**
     * Added attributes to $container element.
     * @default <tt>['data-balance']</tt>
     */
    static get attrs() {
        return ['data_balance'];
    }

    static get _data_balance() {
        return {
            'default': '', 'frozen': 'frozen',
            'ok': 'ok', 'low': 'low', 'neg': 'neg',
        };
    }

    /**
     * <tt>a.balance</tt> with two decimals.
     */
    static prop_balance(a) {
        return a.balance.toFixed(2);
    }

    /**
     * <tt>'COF' if account is cof or 'Non-COF'
     */
    static prop_is_cof(a) {
        return a.is_cof ? 'COF' : 'Non-COF';
    }

    /**
     * Value of data_attribute according to is_frozen status and balance of
     * account <tt>a</tt>.
     */
    static attr_data_balance(a) {
        if      (a.is_frozen)    { return this._data_balance.frozen; }
        else if (a.balance >= 5) { return this._data_balance.ok;  }
        else if (a.balance >= 0) { return this._data_balance.low; }
        else /* a.balance < 0 */ { return this._data_balance.neg; }
    }

}


/**
 * @memberof Formatters
 * @extends Formatters.AccountFormatter
 */
class LIQFormatter extends AccountFormatter {

    /**
     * Rendering a property always returns the empty string
     * @default <tt>''</tt>
     */
    static get_prop() {
        return '';
    }

    /**
     * Attribute <tt>data_balance</tt> is always <tt>ok</tt>.
     */
    static attr_data_balance(a) { return this._data_balance.ok; }

}


/**
 * @memberof Formatters
 * @extends Formatters.Formatter
 */
class CheckoutFormatter extends Formatter {

    /**
     * Properties renderable to html.
     * @default {@link Models.Checkout.props}
     */
    static get props() {
        return Checkout.props;
    }

    /**
     * <tt>c.balance</tt> with two decimals.
     */
    static prop_balance(c) {
        return c.balance.toFixed(2);
    }

}


/**
 * @memberof Formatters
 * @extends Formatters.Formatter
 */

class StatementFormatter extends Formatter {

    /**
     * Properties renderable to html.
     * @default {@link Models.Statement.props}
     */
    static get props() {
        return Statement.props;
    }

    /**
     * <tt>s.balance_old</tt> with two decimals.
     */
    static prop_balance_old(s) {
        return s.balance_old.toFixed(2);
    }

    /**
     * <tt>s.balance_new</tt> with two decimals.
     */
    static prop_balance_new(s) {
        return s.balance_new.toFixed(2);
    }

    /**
     * <tt>s.at</tt> formatted as <tt>DD/MM/YY à HH:MM</tt>.
     */
    static prop_at(s) {
        return moment.isMoment(s.at) ? s.at.tz('Europe/Paris').format('DD/MM/YY à HH:mm') : s.at;
    }

}
