/**
 * @file K-Psul JS
 * @copyright 2017 cof-geek
 * @license MIT
 */


class KPsulManager {

    constructor(env) {
        this._env = env;
        this.account_manager = new AccountManager(this);
        this.checkout_manager = new CheckoutManager(this);
    }

    reset(soft) {
        soft = soft || false;

        this.account_manager.reset();

        if (!soft) {
            this.checkout_manager.reset();
        }
    }

}


class AccountManager {

    constructor() {
        this._$container = $('#account');

        this.account   = new Account();
        this.selection = new AccountSelection(this);
        this.search    = new AccountSearch(this);

        // buttons: search, read or create
        this.$buttons_container = this._$container.find('.buttons');
        this.buttons_templates = {
            create: template`<a href="${'url'}" class="btn btn-primary" target="_blank" title="Créer ce compte"><span class="glyphicon glyphicon-plus"></span></a>`,
            read: template`<a href="${'url'}" class="btn btn-primary" target="_blank" title="Détails du compte"><span class="glyphicon glyphicon-info-sign"></span></a>`,
            search: template`<button class="btn btn-primary search" title="Rechercher"><span class="glyphicon glyphicon-search"></span></button>`,
        }

    }

    is_empty() { return this.account.is_empty(); }

    display() {
        this._display_data();
        this._display_buttons();
    }

    _display_data() {
        this.account.display(this._$container, {
            'prefix_prop': '#account-',
        });
    }

    _display_buttons() {
        var buttons;

        if (this.is_empty()) {
            var trigramme = this.selection.get();
            if (trigramme.isValidTri()) {
                var url = Account.url_create(trigramme);
                buttons = this.buttons_templates.create({url: url});
            } else {  /* trigramme input is empty or invalid */
                buttons = this.buttons_templates.search();
            }
        } else {  /* an account is loaded */
            var url = this.account.url_read;
            buttons = this.buttons_templates.read({url: url});
        }

        this.$buttons_container.html(buttons);
    }

    update(trigramme) {
        if (trigramme !== undefined)
            this.selection.set(trigramme);

        trigramme = trigramme || this.selection.get();

        if (trigramme.isValidTri()) {
            this.account.get_by_apipk(trigramme)
                .done( () => this._update_on_success() )
                .fail( () => this.reset_data() );
        } else {
            this.reset_data();
        }
    }

    _update_on_success() {
        $('#id_on_acc').val(this.account.id);
        this.display();

        kpsul._env.articleSelect.focus();
        kpsul._env.updateBasketAmount();
        kpsul._env.updateBasketRel();
    }

    reset() {
        $('#id_on_acc').val(0);
        this.selection.reset();
        this.search.reset();
        this.reset_data();
    }

    reset_data() {
        this.account.clear();
        this.display();
    }

    focus() {
        this.selection.focus();
        return this;
    }

}


class AccountSelection {

    constructor(manager) {
        this.manager = manager;
        this._$input = $('#id_trigramme');

        this._init_events();
    }

    _init_events() {
        var that = this;

        // user change trigramme
        this._$input
            .on('input', () => this.manager.update());

        // LIQ shortcuts
        this._$input
            .on('keydown', function(e) {
                // keys: 13:Enter|40:Arrow-Down
                if (e.keyCode == 13 || e.keyCode == 40)
                    that.manager.update('LIQ');
            });
    }

    get() {
        return this._$input.val().formatTri();
    }

    set(v) {
        this._$input.val(v);
    }

    focus() {
        this._$input.focus();
    }

    reset() {
        this.set('');
    }

}


class AccountSearch {

    constructor(manager) {
        this.manager = manager;

        this._content = '<input type="text" name="q" id="search_autocomplete" autocomplete="off" spellcheck="false" autofocus><div id="account_results"></div>' ;
        this._input = '#search_autocomplete';
        this._results_container = '#account_results';

        this._init_outer_events();
    }

    open() {
        var that = this;
        this._$dialog = $.dialog({
            title: 'Recherche de compte',
            content: this._content,
            backgroundDismiss: true,
            animation: 'top',
            closeAnimation: 'bottom',
            keyboardEnabled: true,
            onOpen: function() {
                that._$input = $(that._input);
                that._$results_container = $(that._results_container);
                that._init_form()
                    ._init_inner_events();
            },
        });
    }

    _init_form() {
        var that = this;

        this._$input.yourlabsAutocomplete({
            url: Urls['kfet.account.search.autocomplete'](),
            minimumCharacters: 2,
            id: 'search_autocomplete',
            choiceSelector: '.choice',
            placeholder: "Chercher un utilisateur K-Fêt",
            container: that._$results_container,
            box: that._$results_container,
            fixPosition: function() {},
        });

        return this;
    }

    _init_outer_events() {
        var that = this;

        /* open search on button click */
        this.manager._$container
            .on('click', '.search', () => this.open());

        /* open search on Ctrl-F */
        this.manager._$container
            .on('keydown', function(e) {
                if (e.which == 70 && e.ctrlKey) {
                    that.open();
                    e.preventDefault() ;
                }
            });
    }

    _init_inner_events() {
        this._$input.bind('selectChoice',
                (e, choice, autocomplete) => this._on_select(e, choice, autocomplete));
        return this;
    }

    _on_select(e, choice, autocomplete) {
        this.manager.update(choice.find('.trigramme').text());
        this.reset();
    }

    reset() {
        if (this._$dialog !== undefined) {
            this._$dialog.close();
        }
    }
}


class CheckoutManager {

    constructor() {
        this._$container = $('#checkout');
        this.display_prefix = '#checkout-';

        this.checkout = new Checkout();
        this.selection = new CheckoutSelection(this);

        this._$laststatement_container = $('#last_statement');
        this.laststatement = new Statement();
        this.laststatement_display_prefix = '#checkout-last_statement_';

        this.$buttons_container = this._$container.find('.buttons');
        this.buttons_templates = {
            read: template`<a class="btn btn-primary" href="${'url'}" title="En savoir plus" target="_blank"><span class="glyphicon glyphicon-info-sign"></span></a>`,
            statement_create: template`<a href="${'url'}" title="Effectuer un relevé" class="btn btn-primary" target="_blank"><span class="glyphicon glyphicon-euro"></span></a>`,
        }
    }

    update(id) {
        if (id !== undefined)
            this.selection.set(id);

        id = id || this.selection.get();

        var api_options = {
            'last_statement': true,
        };

        this.checkout.get_by_apipk(id, api_options)
            .done( (data) => this._update_on_success(data) )
            .fail( () => this.reset_data() );

        if (kpsul.account_manager.is_empty()) {
            kpsul.account_manager.focus();
        } else {
            kpsul._env.articleSelect.focus().select();
        }
    }

    _update_on_success(data) {
        if (data['laststatement'] !== undefined)
            this.laststatement.from(data['laststatement']);

        $('#id_checkout').val(this.checkout.id);
        this.display();

    }

    is_empty() { return this.checkout.is_empty(); }

    display() {
        this._display_data();
        this._display_laststatement();
        this._display_buttons();
    }

    _display_data() {
        this.checkout.display(this._$container, {
            'prefix_prop': this.display_prefix,
        });
    }

    _display_laststatement() {
        if (this.laststatement.is_empty()) {
            this._$laststatement_container.hide();
        } else {
            this.laststatement.display(this._$laststatement_container, {
               'prefix_prop': this.laststatement_display_prefix
            });
            this._$laststatement_container.show();
        }
    }

    _display_buttons() {
        var buttons = '';
        if (!this.is_empty()) {
            var url_newcheckout = Statement.url_create(this.checkout.id);
            buttons += this.buttons_templates.statement_create({
                url: url_newcheckout});
            var url_read = this.checkout.url_read;
            buttons += this.buttons_templates.read({url: url_read});
        }
        this.$buttons_container.html(buttons);
    }

    reset() {
        $('#id_checkout').val(0);

        this.selection.reset();
        this.reset_data();

        if (this.selection.choices.length == 1)
            this.update(this.selection.choices[0]);
    }

    reset_data() {
        this.checkout.clear();
        this.laststatement.clear();

        this.display();
    }
}


class CheckoutSelection {

    constructor(manager) {
        this.manager = manager;
        this._$input = $('#id_checkout_select');

        this._init_events();

        this.choices = this._$input.find('option[value!=""]').toArray()
            .map(function(opt) {
                return parseInt($(opt).attr('value'));
            });
    }

    _init_events() {
        this._$input.on('change', () => this.manager.update());
    }

    get() {
        return this._$input.val() || 0;
    }

    set(v) {
        this._$input.find('option[value='+ v +']').prop('selected', true);
    }

    reset() {
        this._$input.find('option:first').prop('selected', true);
    }
}

